// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "SceneItemParser.hpp"

using namespace Dawn;

std::vector<std::string> SceneItemParser::getRequiredAttributes() {
  return std::vector<std::string>();
}

std::map<std::string, std::string> SceneItemParser::getOptionalAttributes() {
  return {
    { "ref", "" },
    { "position", "" },
    { "lookAt", "" },
    { "scale", "" },
    { "prefab", "" }
  };
}

int32_t SceneItemParser::onParse(
  Xml *node,
  std::map<std::string, std::string> values,
  struct SceneItem *out,
  std::string *error
) {
  out->ref = values["ref"];

  if(values["position"].size() > 0) {
    out->position = vec3Parser(values["position"], error);
    if(error->size() > 0) return 1;
  }

  if(values["scale"].size() > 0) {
    out->scale = vec3Parser(values["scale"], error);
    if(error->size() > 0) return 1;
  }

  if(values["lookAt"].size() > 0) {
    auto lookAtSplit = stringSplit(values["lookAt"], ",");
    if(lookAtSplit.size() != 6) {
      *error = "Invalid lookAt value: " + values["lookAt"];
      return 1;
    }

    out->lookAtPosition = vec3Parser(lookAtSplit[0] + "," + lookAtSplit[1] + "," + lookAtSplit[2], error);
    if(error->size() > 0) return 1;
    out->lookAtTarget = vec3Parser(lookAtSplit[3] + "," + lookAtSplit[4] + "," + lookAtSplit[5], error);
    if(error->size() > 0) return 1;
  }

  out->prefab = values["prefab"];

  auto itChildren = node->children.begin();
  while(itChildren != node->children.end()) {
    // Parse child nodes, they may be components or not
    auto c = *itChildren;
  
    if(c->node == "child" || c->node == "item") {
      struct SceneItem child;
      child.registry = out->registry;
      auto ret = (SceneItemParser()).parse(c, &child, error);
      if(ret != 0) return ret;
      out->children.push_back(child);
      
    } else if(c->node == "asset") {
      struct SceneAsset asset;
      auto ret = (SceneAssetParser()).parse(c, &asset, error);
      if(ret != 0) return ret;
      out->assets.push_back(asset);
      
    } else {
      struct SceneItemComponent component;
      component.registry = out->registry;
      auto ret = (SceneItemComponentParser()).parse(c, &component, error);
      if(ret != 0) return ret;
      out->components.push_back(component);
    }
    ++itChildren;
  }

  return 0;
}
